"""Communications with MonoX devices. -Adam Outler -
email: adamoutler(a)hackedyour.info
The development environment is Visual Studio Code.
See launch.json for auto-config.
"""

import asyncio
import concurrent.futures
import logging
from datetime import datetime
import os
import select
import sys
from queue import Empty

from socket import AF_INET, SOCK_STREAM, socket
import tempfile
from typing import Iterable

from uart_wifi.errors import ConnectionException

from .response import (
    FileList,
    InvalidResponse,
    MonoXPreviewImage,
    MonoXResponseType,
    MonoXStatus,
    MonoXSysInfo,
)

# Port to listen on

HOST = "192.168.1.254"
COMMAND = "getstatus"
endRequired = ["goprint", "gostop", "gopause", "delfile"]
END = ",end"
ENCODING = "gbk"
_LOGGER = logging.getLogger(__name__)
Any = object()
MAX_REQUEST_TIME = 10  # seconds
Response = Iterable[MonoXResponseType]


class UartWifi:
    """Mono X Class"""

    def __init__(self, ip_address: str, port: int) -> None:
        """Create a communications UartWifi class.
        :ip_address: The IP to initiate communications with.
        :port: The port to use.
        """
        self.server_address = (ip_address, port)
        self.raw = False
        self.telnet_socket = socket(AF_INET, SOCK_STREAM)

    def set_raw(self, raw: bool = True) -> None:
        """Set raw mode.
        :raw: Set to true if we are outputting raw data instead of processed
            classes.
        """
        self.raw = raw

    def send_request(self, message_to_be_sent: str) -> list[MonoXResponseType]:
        """sends the Mono X request.
        :message_to_be_sent: The properly-formatted uart-wifi message as it is
        to be sent.
        :returns: an object from Response class.
        """
        retvalue = "foo"
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:  # 'There is no current event loop...'
            loop = None
        if loop and loop.is_running():
            pool = concurrent.futures.ThreadPoolExecutor()
            result = pool.submit(
                asyncio.run, self._send_request(message_to_be_sent, retvalue)
            ).result()

            return result
        else:
            return_value = asyncio.run(
                self._send_request(message_to_be_sent, retvalue)
            )

        return return_value

    async def _send_request(
        self, message_to_be_sent: str, return_value: list[MonoXResponseType]
    ) -> list[MonoXResponseType]:
        """sends the Mono X request.
        :message_to_be_sent: The properly-formatted uart-wifi message as it is
        to be sent.
        :returns: an object from Response class.
        """
        return_value = await self._async_send_request(message_to_be_sent)
        return return_value

    async def _async_send_request(
        self, message_to_be_sent: str
    ) -> list[MonoXResponseType]:
        """sends the Mono X request.
        :message_to_be_sent: The properly-formatted uart-wifi message as it is
        to be sent.
        :returns: an object from Response class.
        """
        request = bytes(message_to_be_sent, "utf-8")
        received: str = await asyncio.wait_for(
            _do_request(self.telnet_socket, self.server_address, request),
            MAX_REQUEST_TIME,
        )
        if self.raw:
            return received
        processed = _do_handle(received)
        return processed


async def _do_request(
    sock: socket, socket_address: tuple, to_be_sent: bytes
) -> str:
    """Perform the request

    :param sock: the socket to use for the request
    :param request: the request to send
    :return: a MonoX object"""
    text_received = ""
    try:
        sock = _setup_socket(socket_address)
        sock.sendall(to_be_sent)
        sent_string = to_be_sent.decode()
        if sent_string.endswith("shutdown"):
            return "shutdown,end"
        if sent_string.startswith("b'getPreview2"):
            text_received = bytearray()
            print(text_received)
            end_time = datetime.now().microsecond + 1000000
            while (
                not str(text_received).endswith(END)
                and datetime.now().microsecond < end_time
            ):
                text_received.extend(sock.recv(1))
        else:

            read_list = [sock]
            port_read_delay = datetime.now().microsecond
            end_time = datetime.now().microsecond + 500000
            text_received = handle_request(
                sock, text_received, end_time, read_list, port_read_delay
            )

    except (
        OSError,
        ConnectionRefusedError,
        ConnectionResetError,
    ) as exception:
        raise ConnectionException(
            "Could not connect to AnyCubic printer at " + socket_address[0]
        ) from exception
    finally:
        sock.close()
    return text_received


def handle_request(
    sock, text_received, end_time, read_list, port_read_delay
) -> str:
    """performs the request handling"""
    while True:
        current_time = datetime.now().microsecond
        if end_time > current_time or port_read_delay > current_time:
            readable, [], [] = select.select(read_list, [], [])
            for read_port in readable:
                if read_port is sock:
                    port_read_delay = datetime.now().microsecond + 10000
                    text_received += str(read_port.recv(1).decode())
        if text_received.endswith(",end"):
            break
    return text_received


def _setup_socket(socket_address):
    """Setup the socket for communication
    socket_address: the tupple consisting of (ip_address, port).
    """
    _LOGGER.debug("connecting to %s", socket_address)
    sock = socket(AF_INET, SOCK_STREAM)
    sock.settimeout(2)
    sock.connect(socket_address)
    sock.settimeout(MAX_REQUEST_TIME)
    return sock


def __do_preview2(received_message: bytearray()):
    """Handles preview by writing to file.
    :received_message: The message, as received from UART wifi
    protocol, to be converted to an image.
    """

    filename = received_message.decode("utf_8").split(",", 3)[1]
    file = tempfile.gettempdir() + os.path.sep + filename + ".bmp"
    print(file)

    output_file = open(file=file, mode="rb")

    width = 240
    height = 168

    file_size = os.path.getsize(file)
    pos_in_image = 0
    max_line_length = width * 2
    slices = []
    for row in range(0, height):
        current_slice = bytearray()
        for current_byte in range(0, max_line_length):
            current_byte = (row * max_line_length) + current_byte
            if file_size >= current_byte:
                current_slice.append(output_file.read(2))
            current_byte += 2
            slices.append(current_slice)

    my_slice = [[]]
    for byte in slices:
        print(type(byte))
        my_slice[pos_in_image] = byte[0]
        pos_in_image += 1
        if pos_in_image > (240 * 2):
            pos_in_image = 0
            current_slice = bytearray
            my_slice.append(current_slice)

    # image = Image.new("RGB", (width, height))
    # file_format = "bmp"  # The file extension of the sourced data
    print(len(my_slice))

    print(my_slice)

    # bytes(byte_array)
    # image.write(output_file,file_format)

    # output_file.close()

    return MonoXPreviewImage("")


def _do_handle(message: str) -> Iterable[MonoXResponseType]:
    """Perform handling of the message received by the request"""
    if message is None:
        return "no response"

    lines = message.split(",end")
    recognized_response: Iterable = list()
    for line in lines:
        fields: list(str) = line.split(",")
        message_type = fields[0]
        if len(fields) is Empty or len(fields) < 2:
            continue
        if message_type == "getstatus":
            recognized_response.append(__do_status(fields))
        elif message_type == "getfile":
            recognized_response.append(__do_files(fields))
        elif message_type == "sysinfo":
            recognized_response.append(__do_sys_info(fields))
        elif message_type == "gethistory":
            recognized_response.append(__do_get_history(fields))
        elif message_type == "doPreview2":
            recognized_response.append(__do_preview2(fields))
        # goprint,49.pwmb,end
        elif message_type in [
            "goprint",
            "gostop",
            "gopause",
            "getmode",
            "getwifi",
        ]:
            recognized_response.append(InvalidResponse(fields[1]))
        else:
            print("unrecognized command: " + message_type, file=sys.stderr)
            print(line, file=sys.stderr)
            if recognized_response is not None:
                recognized_response.append(InvalidResponse(fields[1]))

    return recognized_response


def __do_get_history(fields: Iterable):
    """Handles history processing."""
    items = []
    for field in fields:
        if field in fields[0] or fields[-1]:
            continue
        items.append(field)
    return items


def __do_sys_info(fields: Iterable):
    """Handles system info processing."""
    sys_info = MonoXSysInfo()
    if len(fields) > 2:
        sys_info.model = fields[1]
    if len(fields) > 3:
        sys_info.firmware = fields[2]
    if len(fields) > 4:
        sys_info.serial = fields[3]
    if len(fields) > 5:
        sys_info.wifi = fields[4]
    return sys_info


def __do_files(fields: Iterable):
    """Handles file processing."""
    files = FileList(fields)
    return files


def __do_status(fields: Iterable):
    """Handles status processing."""
    status = MonoXStatus(fields)
    return status



""""Class to handle printer simulation"""
import select
import socket
import threading
import time


class AnycubicSimulator:
    """ "Simulator for Anycubic Printer."""

    port = "6000"
    printing = False
    serial = "0000170300020034"
    shutdown_signal = False

    def __init__(self, the_ip: str, the_port: int) -> None:
        """Construct the Anycubic Simulator
        :the_ip: The IP address to use internally for opening the port.
            eg. 127.0.0.1, or 0.0.0.0
        :the_port: The port to monitor for responses.
        """
        self.host = the_ip
        self.port = the_port
        self.printing = False
        self.serial = "234234234"

    def sysinfo(self) -> str:
        """return sysinfo type"""
        return "sysinfo,Photon Mono X 6K,V0.2.2," + self.serial + ",SkyNet,end"

    def getfile(self) -> str:
        """return getfile type"""
        return "getfile,Widget.pwmb/0.pwmb,end"

    def getstatus(self) -> str:
        """return getstatus type"""
        if self.printing:
            return (
                "getstatus,print,Widget.pwmb"
                "/46.pwmb,2338,88,2062,51744,6844,~178mL,UV,39.38,0.05,0,end"
            )
        return "getstatus,stop\r\n,end"

    def goprint(self) -> str:
        """Do printing"""
        if self.printing:
            return "goprint,ERROR1,end"
        self.printing = True
        return "goprint,OK,end"

    def gostop(self) -> str:
        """Do Stop printing"""
        if not self.printing:
            return "gostop,ERROR1,end"
        self.printing = False
        return "gostop,OK,end"

    def start_server(self):
        """Start the uart_wifi simualtor server"""
        my_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        my_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        my_socket.bind((self.host, self.port))
        self.port = my_socket.getsockname()[1]
        print(f"Starting printer on {self.host}:{self.port}")
        my_socket.listen(1)
        my_socket.setblocking(False)
        read_list = [my_socket]
        while not AnycubicSimulator.shutdown_signal:
            readable, [], [] = select.select(read_list, [], [])
            for the_socket in readable:
                if the_socket is my_socket:
                    try:
                        conn, addr = the_socket.accept()
                        thread = threading.Thread(
                            target=self.response_selector,
                            args=(conn, addr),
                        )
                        thread.setDaemon(True)
                        thread.start()
                    except Exception:  # pylint: disable=broad-except
                        pass
                    finally:
                        time.sleep(1)

    def response_selector(self, conn: socket.socket, addr) -> None:
        """The connection handler
        :conn: The connection to use
        :addr: address tuple for ip and port
        """
        print(f"Simulator: accepted connection to {addr}")
        decoded_data = ""
        with conn:
            while (
                "," not in decoded_data
                and "\n" not in decoded_data
                and not AnycubicSimulator.shutdown_signal
            ):
                data = conn.recv(1)
                decoded_data += data.decode()
                if "111\n" in decoded_data:
                    decoded_data = ""
                    continue
            try:
                print("Hex:")
                print(" ".join(f"{hex:02x}" for hex in decoded_data.encode()))
                print("Data:")
                print(decoded_data)
            except UnicodeDecodeError:
                pass
            self.send_response(conn, decoded_data)
            decoded_data = ""

    def send_response(self, conn: socket.socket, decoded_data: str) -> None:
        """Send a response

        :conn: The connection to use
        :addr: address tuple for ip and port
        """
        split_data = decoded_data.split(",")
        for split in split_data:
            if split == "":
                continue
            if "getstatus" in split:
                return_value = self.getstatus()
                conn.sendall(return_value.encode())
            if "sysinfo" in split:
                conn.sendall(self.sysinfo().encode())
            if "getfile" in split:
                conn.sendall(self.getfile().encode())
            if "goprint" in split:
                conn.sendall(self.goprint().encode())
                decoded_data = ""
            if "gostop" in split:
                value = self.gostop()
                print("sent:" + value)
                conn.sendall(value.encode())
            if "getmode" in split:
                value = "getmode,0,end"
                print("sent:" + value)
                conn.sendall(value.encode())
                decoded_data = ""
            if "incomplete" in split:
                value = "getmode,0,"
                print("sent:" + value)
                conn.sendall(value.encode())
                decoded_data = ""
            if "multi" in split:
                value = self.getstatus() + self.sysinfo() + "getmode,0,end"
                print("sent:" + value)
                conn.sendall(value.encode())
                decoded_data = ""
            if decoded_data.endswith("shutdown,"):
                value = "shutdown,end"
                print("sent:" + value)
                conn.sendall(value.encode())
                AnycubicSimulator.shutdown_signal = True



#! python3
"""Uart wifi"""

import getopt
import sys
import time
from typing import Iterable

from uart_wifi.communication import UartWifi
from uart_wifi.errors import ConnectionException
from uart_wifi.response import MonoXResponseType


PORT = 6000
HELP = (
    __file__
    + """ | Adam Outler (monox@hackedyour.info) | GPLv3

Usage: monox.py -i <ip address> -c <command>
args:
 -i [--ipaddress=] - The IP address which your Anycubic Mono X can be reached

 -c [--command=] - The command to send.
   Commands may be used one-at-a-time. Only one command may be sent and it is
        expected to be in the format below.
    Command: getstatus - Returns a list of printer statuses.
    Command: getfile - returns a list of files in format <internal name>:
        <file name>.
        When referring to the file via command, use the <internal name>.
    Command: sysinfo - returns Model, Firmware version, Serial Number,
        and wifi network.
    Command: getwifi - displays the current wifi network name.
    Command: gopause - pauses the current print.
    Command: goresume - ends the current print.
    Command: gostop,end - stops the current print.
    Command: delfile,<internal name>,end - deletes a file.
    command: gethistory,end - gets the history and print settings of previous
        prints.
    Command: delhistory,end - deletes printing history.
    Command: goprint,<internal name>,end - Starts a print of the requested file
    Command: getPreview1,<internal name>,end - returns a list of dimensions
        used for the print.

   Not Supported Commands may return unusable results.
    Command (Not Supported): getPreview2,<internal name>,end
     - returns a binary preview image of the print.

   Unknown Commands are at your own risk and experimentation.
   No attempt is made to process or stop execution of these commands.
    Command: detect
    Command: stopUV - unknown
    Command: getpara - unknown
    Command: getmode - unknown
    Command: setname - unknown
    Command: getname - unknown
    Command: setwifi - unknown
    Command: setZero - unknown
    Command: setZhome - unknown
    Command: setZmove - unknown
    Command: setZstop - unknown
    """
)


try:
    opts, args = getopt.gnu_getopt(
        sys.argv, "drhi:c:p:", ["raw", "ipaddress=", "command=", "port="]
    )
# pylint: disable=broad-except
except Exception:
    print(HELP)
    sys.exit(0)

USE_RAW = False
for opt, arg in opts:
    if opt == "-h":
        print(HELP)
        sys.exit()
    elif opt in "-d":
        time.sleep(1)
    elif opt in ("-r", "--raw"):
        USE_RAW = True
    elif opt in ("-i", "--ipaddress"):
        ip_address = arg
    elif opt in ("-p", "--port"):
        PORT = int(arg)
    elif opt in ("-c", "--command"):
        command = arg
        print(arg)

if "ip_address" not in locals():
    print("You must specify the host ip address (-i xxx.xxx.xxx.xxx)")
    sys.exit(1)

if ip_address == "127.0.0.1":
    time.sleep(1)
responses = None  # pylint: disable=invalid-name
# Try 3 times to get the data.
attempts: int = 0
while attempts < 3:
    try:
        uart = UartWifi(ip_address, PORT)
        if USE_RAW:
            uart.raw = True
        responses: Iterable[MonoXResponseType] = uart.send_request(command)

        break
    except ConnectionException:
        attempts += 1


if responses is not None and isinstance(responses, Iterable):
    for response in responses:
        if isinstance(response, MonoXResponseType):
            response.print()
        else:
            print(response)
else:
    print(responses)



#! python3
"""Fake Anycubic Printer for tests"""
import getopt

import sys
from uart_wifi.simulate_printer import AnycubicSimulator


def start_server(the_ip: str, port: int) -> None:
    """Starts the server
    :the_ip: The IP address to use internally for opening the port.
        eg. 127.0.0.1, or 0.0.0.0
    :the_port: The port to monitor for responses.
    """
    AnycubicSimulator(the_ip, int(port)).start_server()


opts, args = getopt.gnu_getopt(sys.argv, "i:p:", ["ipaddress=", "port="])

IP_ADDRESS = "0.0.0.0"
PORT = 6000
for opt, arg in opts:
    if opt in ("-i", "--ipaddress"):
        IP_ADDRESS = arg
    elif opt in ("-p", "--port"):
        PORT = arg
        print("Opening printer on port " + arg)


start_server(IP_ADDRESS, PORT)
